from collections.abc import Sequence
from enum import Enum

from fastapi import HTTPException
from pydantic import BaseModel, RootModel, field_validator
from unitxt.llm_as_judge import EvaluatorNameEnum, EvaluatorTypeEnum, ModelProviderEnum

from .extended_unitxt import ExtendedEvaluatorNameEnum, ExtendedModelProviderEnum
from .judges.types import (
    DirectInstance,
    DirectInstanceResult,
    Instance,
    PairwiseInstance,
    PairwiseInstanceResult,
)
from .model import StoredTestCase


class TaskEnum(str, Enum):
    SUMMARIZATION = "Summarization"
    TEXT_GENERATION = "Text Generation"
    QUESTION_ANSWERING = "Question Answering"


class DomainEnum(str, Enum):
    NEWS_MEDIA_DOMAIN = "News Media"
    HEALTHCARE = "Healthcare"
    ENTERTAINMENT_AND_POP_CULTURE = "Entertainment And Pop Culture"

    SOCIAL_MEDIA = "Social Media"
    CUSTOMER_SUPPORT_AND_BUSSINESS = "Custumer Support And Business"
    GAMING_AND_ENTERTAINMENT = "Gaming And Entertainment"


class PersonaEnum(str, Enum):
    EXPERIENCED_JOURNALIST = "Experienced journalist"
    NOVICE_JOURNALIST = "Novice journalist"
    OPINION_COLUMNIST = "Opinion columnist"
    NEWS_ANCHOR = "News anchor"
    EDITOR = "Editor"

    MEDICAL_RESEARCHER = "Medical researcher"
    GENERAL_PRACTITIONER = "General practitioner"
    PUBLIC_HEALTH_OFFICIAL = "Public health official"
    HEALTH_BLOGGER = "Health blogger"
    MEDICAL_STUDENT = "Medical student"

    FILM_CRITIC = "Film critic"
    CASUAL_SOCIAL_MEDIA_USER = "Casual social media user"
    TABLOID_REPORTER = "Tabloid reporter"
    HARDCORE_FAN_THEORIST = "Hardcore fan/Theorist"
    INFLUENCER_YOUTUBE_REVIEWER = "Inlfuencer/Youtube reviewer"

    INFLUENCER_POSITIVE_BRAND = "Influencer (Positive brand)"
    INTERNET_TROLL = "Internet troll"
    POLITICAL_ACTIVIST = "Political activist (polarizing)"
    BRAND_VOICE = "Brand voice (Corporate social media account)"
    MEMER = "Memer (Meme creator)"
    CUSTOMER_SERVICE_AGENT = "Customer service agent"
    ANGRY_CUSTOMER = "Angry customer"
    CORPORATE_CEO = "Corporate CEO"
    CONSUMER_ADVOCATE = "Consumer advocate"
    MAKETING_SPECIALIST = "Marketing specialist"

    FLAMER = "Flamer (Agressive player)"
    HARDCORE_GAMER = "Hardcore gamer"
    ESPORT_COMENTATOR = "Esport commentator"
    MOVIE_CRITIC = "Movie critic"
    FAN = "Fan (of a TV show, movie, or game)"


class GenerationLengthEnum(str, Enum):
    SHORT = "Short"
    MEDIUM = "Medium"
    LONG = "Long"


class DirectActionTypeEnum(str, Enum):
    REGENERATE = "Regenerate"
    REPHRASE = "Rephrase"
    LONGER = "Elaborate"
    SHORTER = "Shorten"
    CUSTOM = "Custom"


class CriteriaDTO(BaseModel):
    name: str
    description: str
    prediction_field: str
    context_fields: list[str]

    @field_validator("description")
    def validate_criteria(cls, description):
        if len(description.strip()) == 0:
            raise HTTPException(
                status_code=400, detail="Evaluation criteria is required."
            )
        return description


class CriteriaOptionDTO(BaseModel):
    name: str
    description: str


class CriteriaWithOptionsDTO(CriteriaDTO):
    name: str
    description: str
    options: list[CriteriaOptionDTO]


class InstanceDTO(Instance):
    id: str


class DirectInstanceDTO(DirectInstance, InstanceDTO):
    pass


class PairwiseInstanceDTO(PairwiseInstance, InstanceDTO):
    pass


class EvaluationRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: str
    type: EvaluatorTypeEnum
    instances: Sequence[DirectInstanceDTO] | Sequence[PairwiseInstanceDTO]
    criteria: CriteriaDTO | CriteriaWithOptionsDTO


class PairwiseInstanceResultDTO(BaseModel):
    id: str
    result: PairwiseInstanceResult


class PairwiseResultDTO(BaseModel):
    results: list[PairwiseInstanceResultDTO]


class NotebookParams(BaseModel):
    test_case_name: str
    criteria: dict
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum
    provider: ModelProviderEnum
    predictions: list[str | list[str]]
    context_variables: list[dict[str, str]]
    credentials: dict[str, str]
    evaluator_type: EvaluatorTypeEnum
    model_name: str | None = None
    plain_python_script: bool


# class DownloadTestCaseParams(BaseModel):
#     test_case: TestCase


class SyntheticExampleGenerationRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    type: EvaluatorTypeEnum
    criteria: CriteriaWithOptionsDTO | CriteriaDTO
    generation_length: GenerationLengthEnum | None
    task: TaskEnum | None
    domain: DomainEnum | None
    persona: PersonaEnum | None
    per_criteria_option_count: dict[str, int]
    borderline_count: int


class DirectInstanceResultDTO(BaseModel):
    id: str
    result: DirectInstanceResult


class InstanceResultDTO(BaseModel):
    id: str
    result: DirectInstanceResult | PairwiseInstanceResult


class DirectResultDTO(BaseModel):
    results: list[DirectInstanceResultDTO]


class InstanceResult(RootModel):
    root: DirectInstanceResultDTO | PairwiseInstanceResultDTO


class EvaluationResultDTO(BaseModel):
    results: list[InstanceResultDTO]


class TestModelRequestModel(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str


class DirectAIActionRequest(BaseModel):
    action: DirectActionTypeEnum
    selection: str
    text: str
    prompt: str | None = None
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    type: EvaluatorTypeEnum


class DirectAIActionResponse(BaseModel):
    result: str


class FeatureFlagsModel(BaseModel):
    authentication_enabled: bool
    storage_enabled: bool


class PutTestCaseBody(BaseModel):
    user: str
    test_case: StoredTestCase


class DownloadTestCaseBody(BaseModel):
    test_case: StoredTestCase


class DownloadTestDataBody(BaseModel):
    instances: list[DirectInstanceDTO] | list[PairwiseInstanceDTO]
    prediction_field: str


class EvaluatorMetadataAPI(BaseModel):
    name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum | str
    providers: list[ModelProviderEnum | ExtendedModelProviderEnum]


class EvaluatorsResponseModel(BaseModel):
    # model_config = ConfigDict(arbitrary_types_allowed=True)
    evaluators: list[EvaluatorMetadataAPI]


class FixInstanceRequest(BaseModel):
    provider: ModelProviderEnum | ExtendedModelProviderEnum
    llm_provider_credentials: dict[str, str | None]
    evaluator_name: str
    type: EvaluatorTypeEnum
    instance: DirectInstanceDTO
    criteria: CriteriaWithOptionsDTO
    result: DirectInstanceResult


class FixInstanceResponse(BaseModel):
    fixed_response: str



from .base import DirectJudge, Judge, PairwiseJudge, UnitxtInferenceLangchainRunnable
from .const import DEFAULT_JUDGE_INFERENCE_PARAMS
from .dummy_judge import DummyDirectJudge, DummyPairwiseJudge
from .mprometheus_judge import MPrometheusDirectJudge, MPrometheusPairwiseJudge
from .simple_direct_judge import SimpleDirectJudge
from .thesis_antithesis_direct_judge import ThesisAntithesisDirectJudge
from .types import (
    Criteria,
    CriteriaOption,
    DirectInstance,
    DirectInstanceResult,
    DirectPositionalBias,
    Instance,
    PairwiseInstance,
    PairwiseInstanceResult,
    SingleSystemPairwiseResult,
)
from .unitxt_judges import UnitxtDirectJudge, UnitxtPairwiseJudge

__all__: list[str] = [
    "Judge",
    "DummyDirectJudge",
    "DummyPairwiseJudge",
    "SimpleDirectJudge",
    "ThesisAntithesisDirectJudge",
    "UnitxtDirectJudge",
    "UnitxtPairwiseJudge",
    "DirectJudge",
    "PairwiseJudge",
    "Instance",
    "DirectInstance",
    "PairwiseInstance",
    "SingleSystemPairwiseResult",
    "PairwiseInstanceResult",
    "DirectPositionalBias",
    "DirectInstanceResult",
    "DEFAULT_JUDGE_INFERENCE_PARAMS",
    "Criteria",
    "CriteriaOption",
    "UnitxtInferenceLangchainRunnable",
    "MPrometheusDirectJudge",
    "MPrometheusPairwiseJudge",
]



import logging
import os
from pathlib import Path

from .api_types import DomainEnum, GenerationLengthEnum, PersonaEnum

logger = logging.getLogger(__name__)

domain_persona_map = {
    DomainEnum.NEWS_MEDIA_DOMAIN: [
        PersonaEnum.EXPERIENCED_JOURNALIST,
        PersonaEnum.NOVICE_JOURNALIST,
        PersonaEnum.OPINION_COLUMNIST,
        PersonaEnum.NEWS_ANCHOR,
        PersonaEnum.EDITOR,
    ],
    DomainEnum.HEALTHCARE: [
        PersonaEnum.MEDICAL_RESEARCHER,
        PersonaEnum.GENERAL_PRACTITIONER,
        PersonaEnum.PUBLIC_HEALTH_OFFICIAL,
        PersonaEnum.HEALTH_BLOGGER,
        PersonaEnum.MEDICAL_STUDENT,
    ],
    DomainEnum.ENTERTAINMENT_AND_POP_CULTURE: [
        PersonaEnum.FILM_CRITIC,
        PersonaEnum.CASUAL_SOCIAL_MEDIA_USER,
        PersonaEnum.TABLOID_REPORTER,
        PersonaEnum.HARDCORE_FAN_THEORIST,
        PersonaEnum.INFLUENCER_YOUTUBE_REVIEWER,
    ],
    DomainEnum.SOCIAL_MEDIA: [
        PersonaEnum.INFLUENCER_POSITIVE_BRAND,
        PersonaEnum.INTERNET_TROLL,
        PersonaEnum.POLITICAL_ACTIVIST,
        PersonaEnum.BRAND_VOICE,
        PersonaEnum.MEMER,
    ],
    DomainEnum.CUSTOMER_SUPPORT_AND_BUSSINESS: [
        PersonaEnum.CUSTOMER_SERVICE_AGENT,
        PersonaEnum.ANGRY_CUSTOMER,
        PersonaEnum.CORPORATE_CEO,
        PersonaEnum.CONSUMER_ADVOCATE,
        PersonaEnum.MAKETING_SPECIALIST,
    ],
    DomainEnum.GAMING_AND_ENTERTAINMENT: [
        PersonaEnum.FLAMER,
        PersonaEnum.HARDCORE_GAMER,
        PersonaEnum.ESPORT_COMENTATOR,
        PersonaEnum.MOVIE_CRITIC,
        PersonaEnum.FAN,
    ],
}

generation_length_to_sentence_count = {
    GenerationLengthEnum.SHORT: "1-2 sentences",
    GenerationLengthEnum.MEDIUM: "3-5 sentences",
    GenerationLengthEnum.LONG: "5-9 sentences",
}


EVAL_ASSIST_DIR = Path(__file__).parent
STATIC_DIR = Path(os.getenv("STATIC_DIR", EVAL_ASSIST_DIR / "static"))
DATA_DIR = Path(os.getenv("DATA_DIR", EVAL_ASSIST_DIR / "data")).expanduser()
DATA_DIR.mkdir(parents=True, exist_ok=True)
DEFAULT_DATABASE_URL = f"sqlite:////{DATA_DIR / 'evalassist.db'}"
DATABASE_URL = os.getenv("DATABASE_URL", DEFAULT_DATABASE_URL)
TEMPORARY_FILES_FOLDER = EVAL_ASSIST_DIR / "temporary_files"
TEMPORARY_FILES_FOLDER.mkdir(parents=True, exist_ok=True)

DEFAULT_UNITXT_INFERENCE_ENGINE_CACHE_PATH = DATA_DIR / "inference_engine_cache"

UNITXT_INFERENCE_ENGINE_CACHE_PATH = os.getenv("UNITXT_INFERENCE_ENGINE_CACHE_PATH")
if UNITXT_INFERENCE_ENGINE_CACHE_PATH is None:
    UNITXT_INFERENCE_ENGINE_CACHE_PATH = DEFAULT_UNITXT_INFERENCE_ENGINE_CACHE_PATH
    os.environ["UNITXT_INFERENCE_ENGINE_CACHE_PATH"] = str(
        DEFAULT_UNITXT_INFERENCE_ENGINE_CACHE_PATH
    )
else:
    UNITXT_INFERENCE_ENGINE_CACHE_PATH = Path(UNITXT_INFERENCE_ENGINE_CACHE_PATH)
UNITXT_INFERENCE_ENGINE_CACHE_PATH.mkdir(parents=True, exist_ok=True)

UNITXT_CACHE_ENABLED = os.getenv("UNITXT_CACHE_ENABLED", "true").lower() == "true"
STORAGE_ENABLED = os.getenv("STORAGE_ENABLED", "true").lower() == "true"
os.environ["STORAGE_ENABLED"] = str(STORAGE_ENABLED)

AUTHENTICATION_ENABLED = os.getenv("AUTHENTICATION_ENABLED", "false").lower() == "true"

UVICORN_WORKERS = os.getenv("UVICORN_WORKERS", "1")
try:
    UVICORN_WORKERS = int(UVICORN_WORKERS)
    if UVICORN_WORKERS < 1:
        UVICORN_WORKERS = 1
except ValueError:
    UVICORN_WORKERS = 1
    logger.info(f"Invalid UVICORN_WORKERS value, defaulting to {UVICORN_WORKERS}")

CUSTOM_MODELS_PATH = os.getenv(
    "CUSTOM_MODELS_PATH", EVAL_ASSIST_DIR / "custom_models.json"
)

logger.debug(f"EVAL_ASSIST_DIR: {EVAL_ASSIST_DIR}")
logger.debug(f"DATA_DIR: {DATA_DIR}")
if DATABASE_URL.startswith("sqlite"):
    logger.debug(f"DATABASE_URL: {DATABASE_URL}")
logger.debug(
    f"UNITXT_INFERENCE_ENGINE_CACHE_PATH: {UNITXT_INFERENCE_ENGINE_CACHE_PATH}"
)
logger.debug(f"UNITXT_CACHE_ENABLED: {UNITXT_CACHE_ENABLED}")
logger.debug(f"STORAGE_ENABLED: {STORAGE_ENABLED}")
logger.debug(f"AUTHENTICATION_ENABLED: {AUTHENTICATION_ENABLED}")
logger.debug(f"UVICORN_WORKERS: {UVICORN_WORKERS}")
logger.debug(f"CUSTOM_MODELS_PATH: {CUSTOM_MODELS_PATH}")


DIRECT_ACTION_PARAMS = {
    "use_cache": False,
    "seed": None,
    "max_tokens": 200,
    "temperature": 0.7,
}

SYNTHETIC_DATA_GENERATION_PARAMS = {
    "use_cache": False,
    "seed": None,
    "max_tokens": 1200,
    "temperature": 1.0,
    "top_p": 0.9,
    "frequency_penalty": 1.0,
    "presence_penalty": 1.5,
}



import os
import sqlite3
from concurrent.futures import ThreadPoolExecutor, as_completed

import pandas as pd
import requests
from unitxt.inference import CrossProviderInferenceEngine
from unitxt.settings_utils import get_constants

from ..const import EVAL_ASSIST_DIR
from ..judges import DirectJudge


def folder_exists_in_github_repo(owner, repo, folder_path, branch="main"):
    """
    Check if a folder exists in a GitHub repo.

    Parameters:
        owner (str): GitHub username or organization
        repo (str): Repository name
        folder_path (str): Path to folder in the repo (relative to root)
        branch (str): Branch name (default: 'main')

    Returns:
        bool: True if folder exists, False otherwise
    """
    url = f"https://api.github.com/repos/{owner}/{repo}/contents/{folder_path}?ref={branch}"
    try:
        response = requests.get(url, timeout=5)
        if response.status_code == 200:
            # Make sure it's a directory
            items = response.json()
            return isinstance(items, list)
        else:
            return False
    except (
        requests.exceptions.ReadTimeout,
        requests.exceptions.Timeout,
        requests.exceptions.RequestException,
    ):
        return False


def add_tag_to_result(results, keyword, tag_or_tags):
    for benchmark_id in results.keys():
        if keyword in results[benchmark_id]["display_name"]:
            if isinstance(tag_or_tags, list):
                results[benchmark_id]["tags"].extend(tag_or_tags)
            else:
                results[benchmark_id]["tags"].append(tag_or_tags)


def add_url_to_result(results, keyword, url):
    for k in results.keys():
        if keyword in results[k]["display_name"]:
            results[k]["url"] = url


def get_judgebench_readme_url(dataset_name):
    exists = folder_exists_in_github_repo(
        "dmg-illc", "JUDGE-BENCH", f"data/{dataset_name}", "master"
    )
    readme_url = f"https://github.com/dmg-illc/JUDGE-BENCH/blob/master/data/{dataset_name}/README.md"
    return exists, readme_url


def add_judgebench_readme_url(benchmark_id):
    dataset_name = benchmark_id.split("/")[1]
    futures = []
    with ThreadPoolExecutor(2) as executor:
        for option in [dataset_name, dataset_name.replace("_", "-")]:
            futures.append(executor.submit(get_judgebench_readme_url, option))
    for future in as_completed(futures):
        exists, readme_url = future.result()
        if exists:
            return benchmark_id, readme_url
    return benchmark_id, None


def add_judgebench_readme_urls(results):
    futures = []
    with ThreadPoolExecutor(max_workers=20) as executor:
        for benchmark_id in results.keys():
            futures.append(executor.submit(add_judgebench_readme_url, benchmark_id))
    for future in as_completed(futures):
        benchmark_id, readme_url = future.result()
        results[benchmark_id]["url"] = readme_url


def get_judgebench_cards():
    constants = get_constants()
    if constants is None:
        raise ValueError("Error getting unitxt cards: constanst is None")
    judgebench_dir = os.path.join(
        constants.catalog_dir,  # ignore type
        "cards",
        "judge_bench",
    )

    judgebench_cards = []

    for dirpath, _, filenames in os.walk(judgebench_dir):
        for file in filenames:
            if file.endswith(".json"):
                # Get the relative path without the .json extension
                relative_path = os.path.relpath(
                    os.path.join(dirpath, file), judgebench_dir
                )
                without_extension = os.path.splitext(relative_path)[0]
                dotted_path = without_extension.replace(os.path.sep, ".")
                judgebench_cards.append(f"cards.judge_bench.{dotted_path}")

    return judgebench_cards


def get_judge_from_config(
    kwargs: tuple[type[DirectJudge], dict, dict, str],
    inference_engines={},
) -> DirectJudge:
    judge_klass, judge_kwargs, inference_engine_kwargs, model = kwargs
    temperature = (
        inference_engine_kwargs["temperature"]
        if "temperature" in inference_engine_kwargs
        else 0.0
    )
    key = f"{model}{str(temperature)}"
    if key in inference_engines:
        inference_engine = inference_engines[key]
    else:
        inference_engine = CrossProviderInferenceEngine(
            model=model,
            provider="rits",
            temperature=temperature,
            max_tokens=2048,
            cache_batch_size=50,
            data_classification_policy=["public"],
        )
        inference_engines[key] = inference_engine

    return judge_klass(
        inference_engine=inference_engine,
        **judge_kwargs,
    )


def save_evaluation_backup_to_sqlite(result_backup: list[dict]):
    db_path = EVAL_ASSIST_DIR / "benchmark" / "results_backup.db"
    conn = sqlite3.connect(db_path, timeout=10)
    cur = conn.cursor()

    # Create table if it doesn’t exist
    cur.execute("""
    CREATE TABLE IF NOT EXISTS results_backup (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        judge TEXT NOT NULL,
        model TEXT,
        card TEXT NOT NULL,
        sample_index INTEGER NOT NULL,
        instance TEXT,
        result TEXT,
        prediction REAL,
        ground_truth REAL,
        failed BOOLEAN,
        UNIQUE(judge, model, card, sample_index) ON CONFLICT REPLACE
    )
    """)

    cur.executemany(
        """
    INSERT INTO results_backup (
        judge, model, card, sample_index,
        instance, result, prediction, ground_truth, failed
    ) VALUES (
        :judge, :model, :card, :sample_index,
        :instance, :result, :prediction, :ground_truth, :failed
    )
    """,
        result_backup,
    )

    conn.commit()
    conn.close()


def save_results_to_sqlite(results: list[dict]):
    db_path = EVAL_ASSIST_DIR / "benchmark" / "benchmark_results.db"
    conn = sqlite3.connect(db_path, timeout=10)
    cur = conn.cursor()

    # Create table if it doesn’t exist
    cur.execute("""
    CREATE TABLE IF NOT EXISTS benchmark_results (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        judge TEXT NOT NULL,
        model TEXT NOT NULL,
        card TEXT NOT NULL,
        benchmark_name TEXT NOT NULL,
        dataset_name TEXT NOT NULL,
        results TEXT NOT NULL,
        dataset_len TEXT NOT NULL,
        group_by_field TEXT,
        group_by_value TEXT,
        UNIQUE(judge, model, card, group_by_field, group_by_value) ON CONFLICT REPLACE
    )
    """)

    cur.executemany(
        """
    INSERT INTO benchmark_results (
        judge,
        model,
        card,
        benchmark_name,
        dataset_name,
        results,
        dataset_len,
        group_by_field,
        group_by_value
    ) VALUES (
        :judge,
        :model,
        :card,
        :benchmark_name,
        :dataset_name,
        :results,
        :dataset_len,
        :group_by_field,
        :group_by_value
    )
    """,
        results,
    )

    conn.commit()
    conn.close()


def is_result_available(judge, model, card):
    db_path = EVAL_ASSIST_DIR / "benchmark" / "benchmark_results.db"
    conn = sqlite3.connect(db_path, timeout=10)
    cursor = conn.cursor()

    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='benchmark_results'"
    )
    table_exists = cursor.fetchone() is not None

    if table_exists:
        cursor.execute(
            "SELECT EXISTS (SELECT 1 FROM benchmark_results WHERE judge=? AND model=? AND card=?)",
            (judge, model, card),
        )

        exists = cursor.fetchone()[0]  # 1 if exists, 0 if not
        return bool(exists)

    return False


def get_benchmark_results_as_df() -> pd.DataFrame:
    db_path = EVAL_ASSIST_DIR / "benchmark" / "benchmark_results.db"
    conn = sqlite3.connect(db_path, timeout=10)
    return pd.read_sql_query("SELECT * FROM benchmark_results", conn)



import json
import re
from abc import ABC, abstractmethod
from typing import Literal

import nbformat as nbf

from .api_types import NotebookParams
from .judges.const import DEFAULT_JUDGE_INFERENCE_PARAMS
from .utils import get_cross_inference_engine_params


class Cell:
    type: Literal["code", "md"]
    content: str

    def __init__(self, type: Literal["code", "md"], content: str):
        self.type = type
        self.content = content


class EvaluationNotebookGenerator(ABC):
    def __init__(self, params: NotebookParams):
        self.params = params
        self.inference_engine_params = get_cross_inference_engine_params(
            credentials=params.credentials,
            provider=params.provider,
            model_name=params.model_name,
            custom_params=DEFAULT_JUDGE_INFERENCE_PARAMS,
        )
        self.input_fields = {k: "str" for k in params.context_variables[0].keys()}
        self.context_fields = list(params.context_variables[0].keys())
        self.cells: list[Cell] = []

    def generate_notebook(self):
        nb = nbf.v4.new_notebook()
        self._add_title()
        self._add_imports()
        self._add_load_dataset()
        self._add_load_criteria()
        self._add_setup()
        self._add_evaluation()
        for cell in self.cells:
            if cell.type == "md":
                nb.cells.append(nbf.v4.new_markdown_cell(cell.content))
            else:
                nb.cells.append(nbf.v4.new_code_cell(cell.content))
        return nb

    def _add_title(self):
        title = f"# Unitxt {self.get_evaluation_type()} evaluation notebook: {self.params.test_case_name}\n\n"
        title += (
            f'This notebook was generated automatically from your EvalAssist test case "{self.params.test_case_name}". '
            "It contains code to evaluate a set of responses using the specified criteria and evaluator. "
            "EvalAssist uses [unitxt](https://www.unitxt.ai/en/latest/index.html) to create and run the evaluations. "
            "You can find the documentation [here](https://www.unitxt.ai/en/latest/docs/llm_as_judge.html).\n\n"
        )
        self.cells.append(Cell(type="md", content=title))

    def _add_imports(self):
        import_md = "### Import the necessary libraries"
        import_code = self.get_import_code()
        self.cells.append(Cell(type="md", content=import_md))
        self.cells.append(Cell(type="code", content=import_code))

    def _add_load_dataset(self):
        self.cells.append(Cell(type="md", content=self.get_load_dataset_md()))
        self.cells.append(Cell(type="code", content=self.get_load_dataset_code()))

    def _add_load_criteria(self):
        self.cells.append(Cell(type="md", content=self.get_load_criteria_md()))
        self.cells.append(Cell(type="code", content=self.get_load_criteria_code()))

    def _add_setup(self):
        self.cells.append(Cell(type="md", content=self.get_setup_md()))
        self.cells.append(Cell(type="code", content=self.get_setup_code()))

    def _add_evaluation(self):
        self.cells.append(
            Cell(type="md", content="### Evaluate the responses and print the results")
        )
        self.cells.append(Cell(type="code", content=self.get_evaluation_code()))

    @abstractmethod
    def get_evaluation_type(self):
        pass

    @abstractmethod
    def get_import_code(self):
        pass

    @abstractmethod
    def get_load_dataset_md(self):
        pass

    @abstractmethod
    def get_load_dataset_code(self):
        pass

    @abstractmethod
    def get_load_criteria_md(self):
        pass

    @abstractmethod
    def get_load_criteria_code(self):
        pass

    @abstractmethod
    def get_setup_code(self):
        pass

    @abstractmethod
    def get_evaluation_code(self):
        pass

    @abstractmethod
    def get_setup_md(self):
        pass


class DirectEvaluationNotebook(EvaluationNotebookGenerator):
    def get_evaluation_type(self):
        return "direct"

    def get_import_code(self):
        return """
import json
from unitxt.api import evaluate, create_dataset
from unitxt.inference import CrossProviderInferenceEngine
from unitxt.llm_as_judge import LLMJudgeDirect, CriteriaWithOptions
from unitxt.task import Task
import pandas as pd
import nest_asyncio
nest_asyncio.apply()
"""

    def get_load_dataset_md(self):
        return "### Loading the dataset\nThis code block creates a dataset from the context variables and the prediction."

    def get_load_dataset_code(self):
        return f"""context_variables = {json.dumps(self.params.context_variables, indent=4)}
predictions = {json.dumps(self.params.predictions, indent=4)}
dataset_rows = [instance_context_variable | {{"prediction": prediction}} for instance_context_variable, prediction in zip(context_variables, predictions)]
df = pd.DataFrame(dataset_rows)
# load a csv if data is stored in a csv file
# df = pd.read_csv(file_path)
"""

    def get_load_criteria_md(self):
        return """### Load the criteria
The criteria in a direct evaluation needs an option map that matches a string to a numerical value. Replace the NaN value of each option with your desire numerical value.
"""

    def get_load_criteria_code(self):
        option_map_string = {
            option["name"]: float("nan") for option in self.params.criteria["options"]
        }
        return f"""
criteria = {json.dumps(self.params.criteria, indent=4)}
criteria["option_map"] = {json.dumps(option_map_string, indent=4)}
criteria = CriteriaWithOptions.from_obj(criteria)
"""

    def get_setup_code(self):
        params = re.sub(
            r"\btrue\b", "True", json.dumps(self.inference_engine_params, indent=4)
        )
        return f"""
inference_engine = CrossProviderInferenceEngine(**{params})
metric = LLMJudgeDirect(
    inference_engine=inference_engine,
    criteria=criteria,
    context_fields={self.context_fields},
    criteria_field="criteria",
)
dataset_content = df.drop(columns=["prediction"]).to_dict(orient="records")
dataset = create_dataset(
    task=Task(
        input_fields={self.input_fields},
        reference_fields={{}},
        prediction_type=str,
        default_template="templates.empty",
        metrics=[metric],
    ),
    test_set=dataset_content,
    split="test")
"""

    def get_setup_md(self):
        return """### Setup the evaluation
This code block creates the evaluator object of class _LLMJudgeDirect_. It then creates a dataset object from the context variables.
"""

    def get_evaluation_code(self):
        return """predictions = df["prediction"].tolist()
results = evaluate(predictions=predictions, data=dataset)
rows = []
for i, result in enumerate(results):
    instance_scores = result['score']['instance']
    criteria = json.loads(instance_scores[f"{instance_scores['score_name']}_criteria"])
    criteria_str = criteria['name'] if criteria['name'] != "" else criteria['description']
    rows.append({
        'prediction': predictions[i],
        'criteria': criteria_str,
        'score': instance_scores['score'],
        'option': instance_scores[f"{instance_scores['score_name']}_selected_option"]})
results_df = pd.DataFrame(rows)
results_df
"""


class PairwiseEvaluationNotebook(EvaluationNotebookGenerator):
    def get_evaluation_type(self):
        return "pairwise"

    def get_import_code(self):
        return """
import json
from typing import List
from unitxt.api import evaluate, load_dataset
from unitxt.inference import CrossProviderInferenceEngine
from unitxt.llm_as_judge import LLMJudgePairwise, Criteria
from unitxt.blocks import Task, TaskCard
from unitxt.loaders import LoadFromDictionary
from unitxt.templates import NullTemplate

import pandas as pd
import nest_asyncio
nest_asyncio.apply()
"""

    def get_load_dataset_md(self):
        return """### Loading the dataset
This code block creates a dataset from the context variables and the prediction. It simulates the sceario where the dataset is loaded from a csv file.

_Note: in a pairwise dataset, each instance is composed by a context, a criteria and a list of responses. Therefore, this dataset is composed by just one instance._
"""

    def get_load_dataset_code(self):
        system_predictions = [
            {f"system_{i + 1}": pred for i, pred in enumerate(instance_preds)}
            for instance_preds in self.params.predictions
        ]
        return f"""context_variables = {json.dumps(self.params.context_variables, indent=4)}
system_predictions = {json.dumps(system_predictions, indent=4)}
dataset_rows = [instance_context_variable | instance_predictions for instance_context_variable, instance_predictions in zip(context_variables, system_predictions)]
df = pd.DataFrame(dataset_rows)
# load a csv if data is stored in a csv file
# df = pd.read_csv(file_path)
"""

    # **Add missing methods here**
    def get_load_criteria_md(self):
        return """### Load the criteria"""

    def get_load_criteria_code(self):
        return f"""
criteria = Criteria.from_obj({json.dumps(self.params.criteria, indent=4)})
"""

    def get_setup_code(self):
        return f"""metric = LLMJudgePairwise(
    inference_engine=CrossProviderInferenceEngine(**{json.dumps(self.inference_engine_params, indent=4)}),
    criteria=criteria,
    context_fields={self.context_fields},
    criteria_field="criteria",
)
dataset_content = df.filter(regex=r"^(?!system_)").to_dict(orient="records")
data = {{"test": dataset_content}}
card = TaskCard(
    loader=LoadFromDictionary(data=data, data_classification_policy=["public"]),
    task=Task(
        input_fields={self.input_fields},
        prediction_type=List[str],
        metrics=[metric],
        reference_fields={{}},
        default_template=NullTemplate(),
    ),
)
dataset = load_dataset(card=card, split="test")
"""

    def get_setup_md(self):
        return """### Setup the evaluation
This code block creates the evaluator object of class _LLMJudgePairwise_. It then creates a dataset object from the context variables.
"""

    def get_evaluation_code(self):
        return """predictions = df.filter(regex=r"^system_\d+$").values.tolist()
results = evaluate(predictions=predictions, data=dataset)
per_instance_scores = []
winners = []
for i, result in enumerate(results):
    instance_scores = result['score']['instance']
    summary = {
        f'system_{i}': {"winrate": instance_scores[f'{i}_winrate'], "ranking": instance_scores[f'{i}_ranking']}
     for i in range(
        1,
        len(predictions[0]) + 1
    )}
    per_instance_scores.append(summary)
    winners.append(next(iter([k for k, v in summary.items() if v['winrate'] == 1])))
df['winner'] = winners
df['result'] = per_instance_scores
df
"""



from abc import ABC, abstractmethod
from typing import Any, cast

from pydantic import BaseModel, Field, RootModel
from unitxt.llm_as_judge import Criteria as UnitxtCriteria
from unitxt.llm_as_judge import CriteriaOption as UnitxtCriteriaOption
from unitxt.llm_as_judge import CriteriaWithOptions as UnitxtCriteriaWithOptions


class Instance(BaseModel, ABC):
    context: dict[str, str] | None
    expected_result: str | None = None
    metadata: dict[str, Any] | None = None

    @abstractmethod
    def get_prediction(self) -> Any: ...  # noqa: E704


class DirectInstance(Instance):
    response: str

    def get_prediction(self):
        return self.response


class PairwiseInstance(Instance):
    responses: list[str]

    def get_prediction(self):
        return self.responses


class SingleSystemPairwiseResult(BaseModel):
    contest_results: list[bool]
    compared_to: list[int]
    explanations: list[str]
    positional_bias: list[bool] | None = None
    certainty: list[float] | None = None
    winrate: float
    ranking: int
    selections: list[str]


class PairwiseInstanceResult(RootModel):
    root: dict[str, SingleSystemPairwiseResult]


class CriteriaOption(BaseModel):
    name: str
    description: str
    score: float | None = None


class Criteria(BaseModel):
    name: str
    description: str
    prediction_field: str | None = None
    context_fields: list[str] | None = None
    options: list[CriteriaOption] = Field(default_factory=list)

    def get_score_from_option(self, option_name: str):
        try:
            return next(iter(o for o in self.options if o.name == option_name)).score
        except StopIteration:
            return None

    def to_unitxt_criteria(self) -> UnitxtCriteria:
        if len(self.options) > 0:
            return UnitxtCriteriaWithOptions(
                name=self.name,
                description=self.description,
                prediction_field=self.prediction_field,
                context_fields=self.context_fields,
                options=[
                    UnitxtCriteriaOption(
                        name=option.name,
                        description=option.description,
                    )
                    for option in self.options
                ],
                option_map={option.name: option.score for option in self.options}
                if all(option.score is not None for option in self.options)
                else None,
            )
        else:
            return UnitxtCriteria(
                prediction_field=self.prediction_field,
                context_fields=self.context_fields,
                name=self.name,
                description=self.description,
            )

    @staticmethod
    def from_unitxt_criteria(unitxt_criteria: UnitxtCriteria):
        res = Criteria(
            name=unitxt_criteria.name,
            description=unitxt_criteria.description,
            prediction_field=unitxt_criteria.prediction_field,
            context_fields=cast(list[str], unitxt_criteria.context_fields),
        )
        if isinstance(unitxt_criteria, UnitxtCriteriaWithOptions):
            res.options = [
                CriteriaOption(
                    name=option.name,
                    description=option.description,
                    score=unitxt_criteria.option_map[option.name]
                    if unitxt_criteria.option_map is not None
                    and option.name in unitxt_criteria.option_map
                    else None,
                )
                for option in unitxt_criteria.options
            ]
        return res


class DirectPositionalBias(BaseModel):
    detected: bool
    result: "DirectInstanceResult | None" = None


class DirectInstanceResult(BaseModel):
    criteria: Criteria
    option: str
    score: float | None = None
    explanation: str
    feedback: str | None = None
    metadata: dict[str, Any] | None = None
    positional_bias: DirectPositionalBias | None = None



from abc import ABC, abstractmethod
from collections import Counter
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeVar, cast

from langchain.output_parsers import (
    OutputFixingParser,
    ResponseSchema,
    StructuredOutputParser,
)
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompt_values import StringPromptValue
from langchain_core.runnables import RunnableLambda
from pydantic import BaseModel
from unitxt.inference import CrossProviderInferenceEngine, InferenceEngine

from .types import (
    Criteria,
    CriteriaOption,
    DirectInstance,
    DirectInstanceResult,
    DirectPositionalBias,
    Instance,
    PairwiseInstance,
    PairwiseInstanceResult,
)

# ----------------------------------------------------------------------
# Core abstract judge definition
# ----------------------------------------------------------------------
InstanceTypeVar = TypeVar("InstanceTypeVar", bound=Instance)
ReturnVarType = TypeVar("ReturnVarType")


@dataclass
class JudgeDescriptor:
    name: str
    eval_type: Literal["direct", "pairwise"]
    inference_engine_id: str

    def __str__(self) -> str:
        return f"{self.name}-{self.eval_type}-{self.inference_engine_id}"


class UnitxtInferenceEngineMixin:
    inference_engine: InferenceEngine

    def __init__(
        self,
        inference_engine,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.inference_engine = inference_engine


class Judge(
    ABC,
    Generic[InstanceTypeVar, ReturnVarType],
    UnitxtInferenceEngineMixin,
):
    """
    Abstract base class for all judges.

    A *judge* evaluates one or more ``Instance`` objects against a set of
    ``Criteria`` and returns a result specific to the concrete implementation.
    """

    use_self_consistency: bool

    def __init__(
        self,
        use_self_consistency: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.use_self_consistency = use_self_consistency

    def get_inference_engine_id(self) -> str:
        """Return the identifier of the underlying inference engine."""
        return "_".join(self.inference_engine.get_engine_id().split("_")[:-1])

    def get_ai_message_from_prompt(
        self, prompt: str, role: Literal["system", "user", "assistant"] = "user"
    ) -> dict[str, str]:
        return {
            "role": role,
            "content": prompt,
        }

    def evaluate(
        self,
        instances: Sequence[InstanceTypeVar] | Sequence[str],
        criteria: Criteria | Sequence[Criteria] | str,
        check_positional_bias: bool = False,
    ) -> Sequence[ReturnVarType]:
        """Run the judge on a batch of instances and return the results."""
        if (
            isinstance(criteria, Sequence)
            and not isinstance(criteria, str)
            and len(criteria) != len(instances)
        ):
            raise ValueError(
                f"The provided criteria list must be equal in length with the instances. {len(criteria)} != {len(instances)}"
            )
        parsed_criteria: Sequence[Criteria] | Sequence[str]
        if isinstance(criteria, str):
            parsed_criteria = [criteria] * len(instances)
        elif isinstance(criteria, Sequence):
            parsed_criteria = criteria
        else:
            parsed_criteria = [criteria] * len(instances)

        parsed_criteria = self._get_parsed_criteria(parsed_criteria)
        parsed_instances = self._get_instances_from_str(instances)

        if self.use_self_consistency:
            parsed_instances = [
                instance for instance in parsed_instances for _ in range(3)
            ]
            parsed_criteria = [
                criterion for criterion in parsed_criteria for _ in range(3)
            ]

        results: Sequence[ReturnVarType] = self._evaluate(
            instances=parsed_instances,
            criteria=parsed_criteria,
            check_positional_bias=check_positional_bias,
        )

        return results

    def __call__(
        self,
        instances: Sequence[InstanceTypeVar] | Sequence[str],
        criteria: Criteria | Sequence[Criteria] | str,
        check_positional_bias: bool = False,
    ) -> Sequence[ReturnVarType]:
        return self.evaluate(
            instances=instances,
            criteria=criteria,
            check_positional_bias=check_positional_bias,
        )

    @abstractmethod
    def _evaluate(
        self,
        instances: Sequence[InstanceTypeVar],
        criteria: Sequence[Criteria],
        check_positional_bias: bool,
    ) -> Sequence[ReturnVarType]: ...

    @abstractmethod
    def _run(
        self,
        instances: Sequence[InstanceTypeVar],
        criteria: Sequence[Criteria],
    ) -> Sequence[ReturnVarType]: ...

    @abstractmethod
    def _get_instances_from_str(
        self, instances: Sequence[InstanceTypeVar] | Sequence[str] | Sequence[list[str]]
    ) -> Sequence[InstanceTypeVar]: ...

    @abstractmethod
    def _get_parsed_criteria(
        self, criteria: Sequence[Criteria] | Sequence[str]
    ) -> Sequence[Criteria]: ...

    @abstractmethod
    def get_predictions(self, instances: Sequence[InstanceTypeVar]) -> Any:
        """Return the raw predictions (e.g., LLM responses) for the given instances."""
        ...

    @abstractmethod
    def get_descriptor(self) -> JudgeDescriptor:
        """Get an object with primary information of the judge"""
        ...

    @abstractmethod
    def get_name(self) -> str:
        """Get the name of the judge"""
        ...

    def __str__(self):
        return str(self.get_descriptor())


# ----------------------------------------------------------------------
# Concrete abstract subclasses for the two main evaluation modes
# ----------------------------------------------------------------------
class DirectJudge(Judge[DirectInstance, DirectInstanceResult], ABC):
    def _evaluate(
        self,
        instances: Sequence[DirectInstance],
        criteria: Sequence[Criteria],
        check_positional_bias: bool = False,
    ) -> Sequence[DirectInstanceResult]:
        if check_positional_bias:
            results: Sequence[DirectInstanceResult] = self._run(
                instances=[*instances, *instances],
                criteria=[
                    *criteria,
                    *[
                        Criteria(
                            name=criterion.name,
                            description=criterion.description,
                            prediction_field=criterion.prediction_field,
                            context_fields=criterion.context_fields,
                            options=list(reversed(criterion.options)),
                        )
                        for criterion in criteria
                    ],
                ],
            )

            results_len: int = len(instances)
            results = [
                DirectInstanceResult(
                    criteria=results[i].criteria,
                    option=results[i].option,
                    explanation=results[i].explanation,
                    feedback=results[i].feedback,
                    metadata=results[i].metadata,
                    positional_bias=DirectPositionalBias(
                        detected=results[i].option != results[i + results_len].option,
                        result=results[results_len + i],
                    ),
                )
                for i in range(results_len)
            ]
        else:
            results = self._run(instances=instances, criteria=criteria)

        # add numeric scores if possible
        for r in results:
            score: float | None = r.criteria.get_score_from_option(r.option)
            if score is None:
                try:
                    # try to use the option name as the numeric score
                    score = float(r.option)
                except (ValueError, TypeError):
                    pass
            r.score = score

        if self.use_self_consistency:
            # apply majority voting for each of the three evaluation
            parsed_results = []
            for i in range(0, len(results), 3):
                selected_options = [results[i].option for j in range(i, i + 3)]
                most_common_option = Counter(selected_options).most_common(1)[0][0]
                index_of_most_common = selected_options.index(most_common_option)
                to_update_result_index = i + index_of_most_common
                results[to_update_result_index].option = most_common_option
                results[to_update_result_index].score = (
                    sum(cast(float, r.score) for r in results[i : i + 3]) / 3
                    if all(r.score is not None for r in results[i : i + 3])
                    else None
                )  # type: ignore
                parsed_results.append(results[to_update_result_index])
            return parsed_results

        return results

    def _get_instances_from_str(
        self, instances: Sequence[DirectInstance] | Sequence[str] | Sequence[list[str]]
    ) -> Sequence[DirectInstance]:
        parsed_instances: Sequence[DirectInstance]
        if isinstance(instances, Sequence) and all(
            isinstance(x, str) for x in instances
        ):
            parsed_instances = cast(
                Sequence[DirectInstance],
                [
                    DirectInstance(
                        context={},
                        expected_result=None,
                        metadata=None,
                        response=i,
                    )
                    for i in cast(Sequence[str], instances)
                ],
            )
        else:
            parsed_instances = cast(Sequence[DirectInstance], instances)
        return parsed_instances

    def _get_parsed_criteria(
        self, criteria: Sequence[Criteria] | Sequence[str]
    ) -> Sequence[Criteria]:
        if isinstance(criteria, Sequence) and all(isinstance(x, str) for x in criteria):
            return [
                Criteria(
                    name="",
                    description=description,
                    options=[
                        CriteriaOption(name="Yes", description="", score=1.0),
                        CriteriaOption(name="No", description="", score=0.0),
                    ],
                    prediction_field="response",
                )
                for description in cast(Sequence[str], criteria)
            ]
        else:
            return [
                Criteria(
                    name=criterion.name,
                    description=criterion.description,
                    options=criterion.options,
                    prediction_field=criterion.prediction_field,
                )
                for criterion in cast(Sequence[Criteria], criteria)
            ]

    @abstractmethod
    def _run(
        self,
        instances: Sequence[DirectInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[DirectInstanceResult]: ...

    def get_predictions(self, instances: Sequence[DirectInstance]) -> list[str]:
        return [i.response for i in instances]

    def get_descriptor(self) -> JudgeDescriptor:
        return JudgeDescriptor(
            self.get_name(), "direct", self.get_inference_engine_id()
        )


class PairwiseJudge(Judge[PairwiseInstance, PairwiseInstanceResult], ABC):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        if self.use_self_consistency:
            raise ValueError(
                "Self consistency is not supported on pairwise comparison judges yet.s"
            )

    def _evaluate(
        self,
        instances: Sequence[PairwiseInstance],
        criteria: Sequence[Criteria],
        check_positional_bias: bool,
    ) -> Sequence[PairwiseInstanceResult]:
        if check_positional_bias:
            results: Sequence[PairwiseInstanceResult] = self._run(
                instances=[
                    *instances,
                    *[
                        PairwiseInstance(
                            context=i.context,
                            expected_result=i.expected_result,
                            metadata=i.metadata,
                            responses=list(reversed(i.responses)),
                        )
                        for i in instances
                    ],
                ],
                criteria=[
                    *criteria,
                    *criteria,
                ],
            )

            results_len: int = int(len(results) / 2)

            for instance_result, positional_bias_instance_result in zip(
                results[:results_len], results[results_len:]
            ):
                responses_count = len(instance_result.root)
                for i, response_result in enumerate(instance_result.root.values()):
                    positional_bias_result_response = list(
                        positional_bias_instance_result.root.values()
                    )[responses_count - i - 1]
                    response_result.positional_bias = [
                        a != b
                        for a, b in zip(
                            response_result.contest_results,
                            reversed(positional_bias_result_response.contest_results),
                        )
                    ]

            return results
        else:
            return self._run(instances=instances, criteria=criteria)

    @abstractmethod
    def _run(
        self,
        instances: Sequence[PairwiseInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[PairwiseInstanceResult]: ...

    def get_predictions(self, instances: Sequence[PairwiseInstance]) -> list[list[str]]:
        return [i.responses for i in instances]

    def get_descriptor(self) -> JudgeDescriptor:
        return JudgeDescriptor(
            self.get_name(), "pairwise", self.get_inference_engine_id()
        )

    def _get_instances_from_str(
        self,
        instances: Sequence[PairwiseInstance] | Sequence[str] | Sequence[list[str]],
    ) -> Sequence[PairwiseInstance]:
        parsed_instances: Sequence[PairwiseInstance]
        if isinstance(instances, Sequence) and all(
            isinstance(x, str) for x in instances
        ):
            parsed_instances = cast(
                Sequence[PairwiseInstance],
                [
                    PairwiseInstance(
                        context={},
                        expected_result=None,
                        metadata=None,
                        responses=i,
                    )
                    for i in cast(Sequence[list[str]], instances)
                ],
            )
        else:
            parsed_instances = cast(Sequence[PairwiseInstance], instances)
        return parsed_instances

    def _get_parsed_criteria(
        self, criteria: Sequence[Criteria] | Sequence[str]
    ) -> Sequence[Criteria]:
        if isinstance(criteria, Sequence) and all(isinstance(x, str) for x in criteria):
            return [
                Criteria(
                    name="",
                    description=description,
                )
                for description in cast(Sequence[str], criteria)
            ]
        else:
            return cast(Sequence[Criteria], criteria)


# ----------------------------------------------------------------------
# Helper mix‑in for judges that use LangChain runnables
# ----------------------------------------------------------------------
class UnitxtInferenceLangchainRunnable(UnitxtInferenceEngineMixin):
    def _get_runnable_lambda(self) -> RunnableLambda[StringPromptValue, str]:
        """
        Create a LangChain ``RunnableLambda`` that forwards the prompt to the
        underlying ``InferenceEngine`` and returns the raw LLM response.

        Returns
        -------
        RunnableLambda[StringPromptValue, str]
            A callable runnable that can be used in LangChain pipelines.
        """

        def llm_invoke(text: StringPromptValue) -> str:
            # Call the custom model here and return the raw text
            response: str = cast(
                str,
                cast(CrossProviderInferenceEngine, self.inference_engine).infer(
                    dataset=[
                        {
                            "source": text.text,
                            "data_classification_policy": ["public"],
                        }
                    ]
                )[0],
            )
            return response

        return RunnableLambda(func=llm_invoke)

    def get_pydantic_output_fixing_parser(
        self, pydantic_object: type[BaseModel]
    ) -> OutputFixingParser[Any]:
        """
        Create an ``OutputFixingParser`` for a given Pydantic model.

        Parameters
        ----------
        pydantic_object : Type[BaseModel]
            The Pydantic model class used to parse the LLM output.

        Returns
        -------
        OutputFixingParser[Any]
            Configured parser with retry logic.
        """
        return OutputFixingParser.from_llm(
            llm=self._get_runnable_lambda(),
            parser=PydanticOutputParser(pydantic_object=pydantic_object),
            max_retries=3,
        )

    def get_structured_output_fixing_parser(
        self, response_schemas: list[ResponseSchema]
    ) -> OutputFixingParser[Any]:
        """
        Create an ``OutputFixingParser`` for a given Pydantic model.

        Parameters
        ----------
        pydantic_object : Type[BaseModel]
            The Pydantic model class used to parse the LLM output.

        Returns
        -------
        OutputFixingParser[Any]
            Configured parser with retry logic.
        """
        return OutputFixingParser.from_llm(
            llm=self._get_runnable_lambda(),
            parser=StructuredOutputParser.from_response_schemas(response_schemas),
            max_retries=3,
        )



from collections.abc import Sequence
from typing import Literal

from .base import DirectJudge, PairwiseJudge
from .types import (
    Criteria,
    DirectInstance,
    DirectInstanceResult,
    PairwiseInstance,
    PairwiseInstanceResult,
    SingleSystemPairwiseResult,
)


class MPrometheusDirectJudge(DirectJudge):
    m_prometheus_model_name: str

    def __init__(self, m_prometheus_b_params: Literal[3, 7, 14]):
        self.m_prometheus_model_name = (
            f"Unbabel/M-Prometheus-{str(m_prometheus_b_params)}B"
        )

    def get_name(self) -> str:
        return "prometheus"

    def _validate_criteria(self, criteria: Sequence[Criteria]):
        for criterion in criteria:
            if len(criterion.options) != 5:
                raise ValueError(
                    "Criteria must be of Likert type (5 options in crescending order) because that is the only rubric supported by Prometheus models in direct assessment evaluations."
                )

    def _validate_instances(self, instances: Sequence[DirectInstance]):
        for instance in instances:
            if instance.context is not None and "instruction" not in instance.context:
                raise ValueError(
                    f'Prometheus models expect an instruction. Include an "instruction" context variable in each instance. Found context variables: {list(instance.context.keys())}'
                )

    def _run(
        self,
        instances: Sequence[DirectInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[DirectInstanceResult]:
        from prometheus_eval import PrometheusEval
        from prometheus_eval.prompts import (
            ABSOLUTE_PROMPT_WO_REF,
            SCORE_RUBRIC_TEMPLATE,
        )
        from prometheus_eval.vllm import VLLM

        self._validate_criteria(criteria)
        self._validate_instances(instances)

        parsed_criteria: list[str] = [
            SCORE_RUBRIC_TEMPLATE.format(
                **{
                    "criteria": f"{criterion.name}: {criterion.description}",
                    **{
                        f"score{i + 1}_description": option.description
                        for i, option in enumerate(criterion.options)
                    },
                }
            )
            for criterion in criteria
        ]

        instructions = [
            instance.context["instruction"] if instance.context is not None else ""
            for instance in instances
        ]
        responses = [instance.response for instance in instances]

        model = VLLM(model=self.m_prometheus_model_name, max_model_len=4096)
        # model = LiteLLM(f"huggingface/{self.m_prometheus_model_name}")
        judge = PrometheusEval(
            model=model, absolute_grade_template=ABSOLUTE_PROMPT_WO_REF
        )

        feedbacks, scores = judge.absolute_grade(
            instructions=instructions,
            responses=responses,
            rubric=parsed_criteria,
        )

        return [
            DirectInstanceResult(
                criteria=criterion,
                option=criterion.options[score - 1].name,
                score=score,
                explanation=feedback,
            )
            for feedback, score, criterion in zip(feedbacks, scores, criteria)
        ]


class MPrometheusPairwiseJudge(PairwiseJudge):
    m_prometheus_model_name: str

    def __init__(self, m_prometheus_b_params: Literal[3, 7, 14]):
        self.m_prometheus_model_name = (
            f"Unbabel/M-Prometheus-{str(m_prometheus_b_params)}B"
        )

    def get_name(self) -> str:
        return "prometheus"

    def _validate_instances(self, instances: Sequence[PairwiseInstance]):
        for instance in instances:
            if instance.context is not None and "instruction" not in instance.context:
                raise ValueError(
                    f'Prometheus models expect an instruction. Include an "instruction" context variable in each instance. Found context variables: {list(instance.context.keys())}'
                )
            if len(instance.responses) != 2:
                raise ValueError(
                    "Prometheus only allows for two responses to be compared. Support for comparing more than two responsens will be supported by EvalAssist soon."
                )

    def _run(
        self,
        instances: Sequence[PairwiseInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[PairwiseInstanceResult]:
        from prometheus_eval import PrometheusEval
        from prometheus_eval.prompts import RELATIVE_PROMPT_WO_REF
        from prometheus_eval.vllm import VLLM

        self._validate_instances(instances)

        instructions = [
            instance.context["instruction"] if instance.context is not None else ""
            for instance in instances
        ]
        responses_A = [instance.responses[0] for instance in instances]
        responses_B = [instance.responses[1] for instance in instances]
        model = VLLM(model=self.m_prometheus_model_name, max_model_len=4096)
        # model = LiteLLM(f"huggingface/{self.m_prometheus_model_name}")
        judge = PrometheusEval(
            model=model, absolute_grade_template=RELATIVE_PROMPT_WO_REF
        )
        parsed_criteria = [
            f"{criterion.name}: {criterion.description}" for criterion in criteria
        ]
        result: tuple[list[str], list[str]] = judge.relative_grade(
            instructions=instructions,
            responses_A=responses_A,
            responses_B=responses_B,
            rubric=parsed_criteria,
        )  # type: ignore

        feedbacks, scores = result

        results: list[PairwiseInstanceResult] = []
        # systems_per_instance = len(instances[0].responses)
        # comparisons_per_instance =  systems_per_instance - 1
        for i, (instance, feedback, score) in enumerate(
            zip(instances, feedbacks, scores)
        ):
            instance_result: dict[str, SingleSystemPairwiseResult] = {}
            instance_result["system_1"] = SingleSystemPairwiseResult(
                contest_results=[score == "A"],
                compared_to=[1],
                explanations=[feedback],
                positional_bias=[False],
                winrate=1.0,
                ranking=1 if score == "A" else 0,
                selections=["1" if score == "A" else "2"],
            )

            instance_result["system_2"] = SingleSystemPairwiseResult(
                contest_results=[score == "B"],
                compared_to=[0],
                explanations=[feedback],
                positional_bias=[False],
                winrate=1.0,
                ranking=1 if score == "B" else 0,
                selections=["1" if score == "A" else "2"],
            )
            results.append(PairwiseInstanceResult(instance_result))
        return results



from itertools import cycle

from evalassist.benchmark import run_benchmarks
from evalassist.judges import DirectJudge, SimpleDirectJudge, UnitxtDirectJudge

if __name__ == "__main__":
    MAX_WORKERS = 7
    BATCH_SIZE = 25
    RITS_API_KEYS = None
    INSTANCES_PER_DATASET = 200
    # List of models to benchmark
    MODELS = [
        # "gpt-oss-120b",
        "llama-3-3-70b-instruct",
        "llama-4-scout",
        "llama-4-maverick",
        "granite-3-3-8b-instruct",
        "deepseek-v3",
        "phi-4",
        "mistral-small-instruct",
    ]
    api_key_cycle = cycle(RITS_API_KEYS if RITS_API_KEYS is not None else [None])

    inference_engines = {}
    JUDGE_CONFIGS: list[tuple[type[DirectJudge], dict, dict]] = [
        # (
        #     SimpleDirectJudge,
        #     {
        #         "generate_synthetic_persona": True,
        #         "use_self_consistency": True,
        #     },
        #     {
        #         'temperature': 0.3,
        #     },
        # ),
        # (
        #     SimpleDirectJudge,
        #     {
        #         "generate_synthetic_persona": False,
        #         "use_self_consistency": True,
        #     },
        #     {
        #         'temperature': 0.3,
        #     },
        # ),
        # (
        #     SimpleDirectJudge,
        #     {
        #         "generate_synthetic_persona": True,
        #         "use_self_consistency": False,
        #     },
        #     {
        #         'temperature': 0.0,
        #     },
        # ),
        (
            SimpleDirectJudge,
            {
                "generate_synthetic_persona": False,
                "use_self_consistency": False,
            },
            {
                "temperature": 0.0,
            },
        ),
        (UnitxtDirectJudge, {}, {}),
        # (ThesisAntithesisDirectJudge, {}, {}),
    ]
    JUDGE_CONFIGS_WITH_MODEL: list[tuple[type[DirectJudge], dict, dict, str]] = []
    for judge_config in JUDGE_CONFIGS:
        for model in MODELS:
            new_tuple = judge_config + (model,)
            JUDGE_CONFIGS_WITH_MODEL.append(new_tuple)

    run_benchmarks(
        judge_configs=JUDGE_CONFIGS_WITH_MODEL,
        max_workers=MAX_WORKERS,
        instances_per_dataset=INSTANCES_PER_DATASET,
        dataset_keyword_filters=["drop", "esnli", "biggen"],
        # dataset_keyword_selectors=["cola"],
    )
    # model = "granite-3-3-8b-instruct"
    # inference_engine = CrossProviderInferenceEngine(
    #     model=model,
    #     provider="rits",
    #     temperature=0,
    #     max_tokens=2048,
    #     data_classification_policy=["public"],
    # )
    # print(run_single_model_card(
    #     "cards.judge_bench.roscoe.overall.drop.overall_quality",
    #     "gpt-oss-120b",
    #     UnitxtDirectJudge(inference_engine)))



import logging
import random
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from textwrap import dedent
from typing import Any, cast

from langchain.output_parsers import OutputFixingParser
from langchain.prompts import PromptTemplate
from langchain_core.exceptions import OutputParserException
from pydantic import BaseModel, Field, field_validator
from unitxt.inference import InferenceEngine

from .base import (
    DirectJudge,
    UnitxtInferenceEngineMixin,
    UnitxtInferenceLangchainRunnable,
)
from .types import Criteria, DirectInstance, DirectInstanceResult

logger = logging.getLogger(__name__)


class SimpleDirectJudge(
    DirectJudge, UnitxtInferenceLangchainRunnable, UnitxtInferenceEngineMixin
):
    generate_synthetic_persona: bool
    generate_feedback: bool
    judge_description_prompt: str | None

    def __init__(
        self,
        inference_engine: InferenceEngine,
        generate_synthetic_persona: bool = False,
        judge_description_prompt: str | None = None,
        generate_feedback: bool = False,
        use_self_consistency: bool = False,
    ):
        super().__init__(
            inference_engine=inference_engine, use_self_consistency=use_self_consistency
        )
        if generate_synthetic_persona and judge_description_prompt:
            raise ValueError(
                "Either provide set generate_synthetic_persona to False or don't provide a judge_description_prompt."
            )
        self.generate_synthetic_persona = generate_synthetic_persona
        self.judge_description_prompt = judge_description_prompt
        self.generate_feedback = generate_feedback

    def parse_response(self, unparsed_response, output_parser, klass, criterion):
        metadata = {}
        try:
            parsed_response = output_parser.parse(completion=unparsed_response)
            metadata["failed_generation"] = False
        except OutputParserException as e:
            logger.debug(
                f"Selected random option for model {self.inference_engine.get_engine_id()} because it was unable to generate a chosen option"
            )
            parsed_response = klass(
                selected_option=random.choice(
                    [o.name for o in criterion.options]  # nosec
                ),
                assessment="",
            )
            metadata["failed_generation"] = True
            metadata["failed_generation_original_output"] = unparsed_response
            metadata["failed_generation_final_output"] = e.llm_output
        return parsed_response, metadata

    def get_name(self) -> str:
        return f"simple{'_with_synthetic_persona' if self.generate_synthetic_persona else ''}{'_with_feedback' if self.generate_feedback else ''}{'_with_self_consistency' if self.use_self_consistency else ''}"

    def generate_personas(
        self,
        context_sections: list[str],
        predictions: list[str],
        criteria: Sequence[Criteria],
    ) -> list[tuple[str, str]]:
        unique_criteria_instance: list[tuple[Criteria, tuple[str, str]]] = list(
            {
                criterion.name: (criterion, (context_section, prediction))
                for criterion, context_section, prediction in zip(
                    criteria, context_sections, predictions
                )
            }.values()
        )
        unique_criteria, instance_examples = zip(*unique_criteria_instance)  # type: ignore
        unique_criteria: list[Criteria] = list(unique_criteria)
        instance_examples: list[tuple[str, str]] = list(instance_examples)

        instance_examples_str = [
            context_section + "\nText to evaluate: " + prediction
            for context_section, prediction in instance_examples
        ]

        synthetic_persona_klasses = []
        output_parsers = []
        format_instructions = []
        for criterion in unique_criteria:

            class SyntheticPersona(BaseModel):
                persona_name: str = Field(
                    default=...,
                    description=f"The name of the persona responsible for evaluating the {criterion.prediction_field} according to the criterion {criterion.name}.",
                )
                persona_description: str = Field(
                    ...,
                    description="The description of why the <persona_name> is ideal to perform the evaluation. Don't include the the initial 'you'. For example: 'an expert on evaluating text based on a rubric' or 'a customer support specialist experienced in clarity and tone assessment'.",
                )

            synthetic_persona_klasses.append(SyntheticPersona)

            output_parser: OutputFixingParser = self.get_pydantic_output_fixing_parser(
                SyntheticPersona
            )
            output_parsers.append(output_parser)
            format_instructions.append(output_parser.get_format_instructions())

        template = PromptTemplate(
            input_variables=[
                "criteria_name_section"
                "criteria_description"
                "criteria_options"
                "prediction_field"
                "instance_example"
                "format_instruction"
            ],
            template=dedent(
                text="""\
                    Your task is to generate a persona that is the most appropriate to evaluate a text based on the following criteria.
                    You will be provided with the criteria name, description and options and an example instance.

                    ### Criterion:

                    {criteria_name_section}
                    Description: {criteria_description}
                    Options:
                    {criteria_options}

                    ### Example instance

                    {instance_example}

                    For the persona, you will generate the name or role (e.g. a doctor, a philosopher, a lawyer) and a brief description that makes emphasis on what makes the persona the ideal for performing the evaluation (e.g. have a lot of experience reading and writing email summaries).

                    ### Output format

                    The persona info will be used as this:
                    "You are <persona_name>. Your task is to evaluate a {prediction_field}. You <persona_description>".

                    {format_instruction}
                """
            ),
        )

        prompts = [
            template.format(
                criteria_name_section=f"Criteria name: {criterion.name}"
                if criterion.name
                else "",
                criteria_description=criterion.description,
                criteria_options="\n".join(
                    [
                        f"- {o.name}{f': {o.description}' if o.description else ''}"
                        for o in criterion.options
                    ]
                ),
                prediction_field=criterion.prediction_field
                if criterion.prediction_field
                else "text",
                instance_example=instance_example_str,
                format_instruction=format_instruction,
            )
            for criterion, instance_example_str, format_instruction in zip(
                criteria, instance_examples_str, format_instructions
            )
        ]

        unparsed_responses = self.inference_engine(
            [
                {"source": prompt, "data_classification_policy": ["public"]}
                for prompt in prompts
            ]
        )

        parsed_responses = [
            output_parser.parse(unparsed_response)
            for output_parser, unparsed_response in zip(
                output_parsers, unparsed_responses
            )
        ]
        personas = [
            (persona.persona_name, persona.persona_description)
            for persona in parsed_responses
        ]  # type: ignore
        criteria_name_to_persona = {
            criterion.name: persona
            for criterion, persona in zip(unique_criteria, personas)
        }
        personas_completed = [
            criteria_name_to_persona[criterion.name] for criterion in criteria
        ]
        return personas_completed

    def _run(
        self,
        instances: Sequence[DirectInstance],
        criteria: Sequence[Criteria],
    ) -> list[DirectInstanceResult]:
        output_parsers: list[OutputFixingParser] = []
        format_instructions_list = []
        criteria_options_list = []
        classes = []
        feedback_step_sections = []
        prediction_fields = []
        for criterion in criteria:
            criterion_names = [o.name for o in criterion.options]

            class DynamicOutputJudgeModel(BaseModel):
                assessment: str = Field(..., description="Step by step assessment")
                selected_option: str = Field(
                    ...,
                    description=f"The chosen option. Any of {', '.join([o.name for o in criterion.options])}",
                )

                @field_validator("selected_option")
                def validate_value(cls, v):
                    if v not in criterion_names:
                        raise ValueError(f"value must be one of {criterion_names}")
                    return v

            class DynamicOutputJudgeModelWithFeedback(DynamicOutputJudgeModel):
                feedback: str = Field(
                    default="",
                    description=f"Actionable suggestions that would help improve the evaluated {criterion.prediction_field if criterion.prediction_field is not None else 'response'} based on the assessment",
                )

            output_model_klass = (
                DynamicOutputJudgeModelWithFeedback
                if self.generate_feedback
                else DynamicOutputJudgeModel
            )

            classes.append(output_model_klass)

            output_parser: OutputFixingParser[DynamicOutputJudgeModel] = (
                self.get_pydantic_output_fixing_parser(output_model_klass)
            )
            output_parsers.append(output_parser)

            format_instructions: str = output_parser.get_format_instructions()
            format_instructions_list.append(format_instructions)

            criteria_options: str = "\n".join(
                [
                    f"- {option.name}{f': {option.description}' if option.description else ''}"
                    for option in criterion.options
                ]
            )
            criteria_options_list.append(criteria_options)

            prediction_field = (
                criterion.prediction_field
                if criterion.prediction_field is not None
                else "response"
            )
            prediction_fields.append(prediction_field)

            feedback_step_section = (
                f'5. At the end, provide "feedback" consisting of actionable suggestions that would help improve the evaluated {prediction_field}. Unlike the assessment, which explains the reasoning behind the judgment, the feedback should focus on guiding refinement. For example, in creative writing, it could suggest improving clarity, coherence, or narrative flow. In analytical tasks, it could recommend strengthening evidence, refining arguments, or correcting inaccuracies. Keep feedback concise and specific enough to support iterative improvement. If you consider that the {prediction_field} is optimal, leave the "feedback" field empty ("")'
                if self.generate_feedback
                else ""
            )
            feedback_step_sections.append(feedback_step_section)

        predictions: list[str] = [i.response for i in instances]
        context_variables_list: list[dict[str, str]] = [
            instance.context if instance.context is not None else {}
            for instance in instances
        ]
        str_context_variables_list: list[str | None] = [
            "\n\n".join(f"- {k}: {v}" for k, v in c.items()) if len(c) else None
            for c in context_variables_list
        ]

        context_sections: list[str] = [
            ("\n\n### Context\n\n" + c + "\n") if c is not None else ""
            for c in str_context_variables_list
        ]
        if self.judge_description_prompt:
            judge_description_sections = [self.judge_description_prompt] * len(criteria)
        else:
            if self.generate_synthetic_persona:
                personas = self.generate_personas(
                    context_sections=context_sections,
                    predictions=predictions,
                    criteria=criteria,
                )
            else:
                persona_name, persona_description = (
                    "an evaluator",
                    "an expert on evaluating text based on a rubric.",
                )
                personas = [(persona_name, persona_description)] * len(criteria)
            judge_description_sections = [
                f"You are a {persona_name}. You are {persona_description}"
                for persona_name, persona_description in personas
            ]

        prompt_template = PromptTemplate(
            input_variables=[
                "text_to_evaluate",
                "context_section",
                "criteria_name_section",
                "criteria_description",
                "criteria_options",
                "format_instructions",
                "prediction_field",
                "feedback_step_section",
                "judge_description_section",
            ],
            template=dedent(
                text="""\
                {judge_description_section}

                You will be given:
                - **Criterion** (name, description, options)
                - **Optional context**
                - **The {prediction_field}** to evaluate

                ### Important steps:

                1. Think step-by‑step through your reasoning about which option best fits.
                2. Write your full chain‑of‑thought *only* inside the `assessment` JSON field.
                3. The chain-of-thought should use markdown code for easier reading and parsing.
                4. Set `"selected_option"` to one of the provided options based on the assessment.
                {feedback_step_section}

                ### Criteria:{criteria_name_section}
                Description: {criteria_description}
                Options:
                {criteria_options}{context_section}

                ### The {prediction_field} to evaluate

                {text_to_evaluate}

                ### Output format

                {format_instructions}
            """,
            ),
        )

        prompts: list[str] = [
            prompt_template.format(
                text_to_evaluate=prediction,
                context_section=context_section,
                criteria_name_section=f"\n\nCriteria name: {criterion.name}"
                if criterion.name
                else "\n",
                criteria_description=criterion.description,
                criteria_options=criterion_options,
                format_instructions=format_instructions,
                prediction_field=prediction_field,
                feedback_step_section=feedback_step_section,
                judge_description_section=judge_description_section,
            )
            for prediction, context_section, criterion, criterion_options, format_instructions, prediction_field, feedback_step_section, judge_description_section in zip(
                predictions,
                context_sections,
                criteria,
                criteria_options_list,
                format_instructions_list,
                prediction_fields,
                feedback_step_sections,
                judge_description_sections,
            )
        ]

        unparsed_responses: list[str] = cast(
            list[str],
            self.inference_engine.infer(
                dataset=[
                    {"source": prompt, "data_classification_policy": ["public"]}
                    for prompt in prompts
                ]
            ),
        )
        parsed_responses: list[Any] = []

        futures = []
        with ThreadPoolExecutor(max_workers=10) as executor:
            for unparsed_response, output_parser, klass, criterion in zip(
                unparsed_responses, output_parsers, classes, criteria
            ):
                futures.append(
                    executor.submit(
                        self.parse_response,
                        unparsed_response,
                        output_parser,
                        klass,
                        criterion,
                    )
                )

        parsed_responses = []
        parsing_metadatas = []
        for future in futures:
            parsed_response, metadata = future.result()
            parsed_responses.append(parsed_response)
            parsing_metadatas.append(metadata)

        explanations: list[str] = [r.assessment for r in parsed_responses]
        selected_options: list[str] = [r.selected_option for r in parsed_responses]
        feedbacks: list[str | None] = [
            None if not self.generate_feedback else r.feedback for r in parsed_responses
        ]
        return [
            DirectInstanceResult(
                criteria=criterion,
                option=selected_option,
                explanation=explanation,
                feedback=feedback,
                positional_bias=None,
                metadata={
                    **parsing_metadata,
                    "prompt": prompt,
                    "unparsed_response": unparsed_response,
                },
            )
            for selected_option, explanation, feedback, prompt, unparsed_response, criterion, parsing_metadata in zip(
                selected_options,
                explanations,
                feedbacks,
                prompts,
                unparsed_responses,
                criteria,
                parsing_metadatas,
            )
        ]



from collections.abc import Sequence

from .base import DirectJudge, PairwiseJudge
from .types import (
    Criteria,
    DirectInstance,
    DirectInstanceResult,
    DirectPositionalBias,
    PairwiseInstance,
    PairwiseInstanceResult,
    SingleSystemPairwiseResult,
)


class DummyDirectJudge(DirectJudge):
    def get_name(self) -> str:
        return "dummy"

    def _run(
        self,
        instances: Sequence[DirectInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[DirectInstanceResult]:
        return [
            DirectInstanceResult(
                criteria=criteria[0],
                option=criteria[0].options[0].name,
                explanation="explanation",
                positional_bias=DirectPositionalBias(
                    detected=False,
                ),
            )
            for _ in range(len(instances))
        ]


class DummyPairwiseJudge(PairwiseJudge):
    def get_name(self) -> str:
        return "dummy"

    def _run(
        self,
        instances: Sequence[PairwiseInstance],
        criteria: Sequence[Criteria],
    ) -> Sequence[PairwiseInstanceResult]:
        results: list[PairwiseInstanceResult] = []
        systems_per_instance = len(instances[0].responses)
        comparisons_per_instance = systems_per_instance - 1
        for i, instance in enumerate(instances):
            instance_result: dict[str, SingleSystemPairwiseResult] = {}
            instance_result[f"system_{i}"] = SingleSystemPairwiseResult(
                contest_results=[True for _ in range(comparisons_per_instance)],
                compared_to=[True for _ in range(comparisons_per_instance)],
                explanations=["Explanations" for _ in range(comparisons_per_instance)],
                positional_bias=[False for _ in range(comparisons_per_instance)],
                winrate=1.0,
                ranking=1,
                selections=["1" for _ in range(comparisons_per_instance)],
            )
            results.append(PairwiseInstanceResult(instance_result))
        return results



import json
import time
from collections.abc import Callable
from datetime import datetime

from evalassist.model import LogRecord
from fastapi import Request, Response
from fastapi.routing import APIRoute
from sqlmodel import Session
from starlette.background import BackgroundTask

from .const import STORAGE_ENABLED
from .database import engine  # Assumes you have engine/session setup

ignored_endpoints = [
    "/health",
    "/evaluators/",
    "/criterias/",
    "/test_case/",
    "/user/",
    "/default-credentials/",
    "/benchmarks/",
    "/domains-and-personas/",
    "/feature-flags/",
    "/version/",
]

ignored_methods_per_endpoint = {"/test_case/": ["GET"]}


def log_info(method, path, req_body, res_body, headers, runtime):
    if not STORAGE_ENABLED:
        return

    if path in ignored_endpoints and (
        path not in ignored_methods_per_endpoint
        or method in ignored_methods_per_endpoint[path]
    ):
        return

    record = {
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "runtime": runtime,
    }

    if path != "/log_user_action/":
        # adapt non log user action endpoint calls
        record["action"] = f"{method} {path.replace('/', '')}"

        record_content = {}
        if req_body:
            req = json.loads(req_body.decode())
            if "llm_provider_credentials" in req:
                req["llm_provider_credentials"] = ""
            try:
                req["test_case"]["content"] = json.loads(req["test_case"]["content"])
            except Exception:  # nosec B110
                pass
            record_content["request"] = req

        if res_body:
            res = json.loads(res_body.decode())
            try:
                res["content"] = json.loads(res["content"])
            except Exception:  # nosec B110
                pass
            record_content["response"] = res

        record["content"] = record_content
    else:
        req = json.loads(req_body.decode())
        record["action"] = req["action"]
        record["content"] = req["content"]

    log_record = LogRecord(
        data=json.dumps(record), user_id=headers.get("user_id", None)
    )

    with Session(engine) as session:
        session.add(log_record)
        session.commit()


class LoggingRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: Request) -> Response:
            req_body = await request.body()
            start_timestamp = time.time()
            response = await original_route_handler(request)
            end_timestamp = time.time()
            runtime = round(end_timestamp - start_timestamp, 2)
            tasks = response.background
            task = BackgroundTask(
                log_info,
                request.method,
                request.url.path,
                req_body,
                response.body if hasattr(response, "body") else None,
                request.headers,
                runtime,
            )

            # check if the original response had background tasks already attached to it
            if tasks:
                tasks.add_task(task)  # add the new task to the tasks list
                response.background = tasks
            else:
                response.background = task

            return response

        return custom_route_handler



from enum import Enum

from unitxt.inference import HFAutoModelInferenceEngine
from unitxt.llm_as_judge_constants import (
    EVALUATOR_TO_MODEL_ID,
    EVALUATORS_METADATA,
    EvaluatorMetadata,
    EvaluatorNameEnum,
    ModelProviderEnum,
)


class ExtendedEvaluatorNameEnum(Enum):
    """This enums adds models that are not present in the original unitxt EvaluatorNameEnum"""

    GRANITE_GUARDIAN3_1_2B = "Granite Guardian 3.1 2b"
    GRANITE_GUARDIAN3_1_8B = "Granite Guardian 3.1 8b"
    GRANITE_GUARDIAN3_2_3B = "Granite Guardian 3.2 3b"
    GRANITE_GUARDIAN3_2_5B = "Granite Guardian 3.2 5b"
    GRANITE_GUARDIAN3_3_8B = "Granite Guardian 3.3 8b"
    LLAMA_3_3_70B_FREE = "Llama 3.3 70B Free"
    DEEPSEEK_R1_DISTILLED_LLAMA_70B_FREE = "DeepSeek R1 Distilled Llama 70B Free"
    PHI4 = "Phi-4"
    MIXTRAL_SMALL = "Mixtral Small"
    MIXTRAL_MEDIUM = "Mixtral Medium"
    GPT_OSS_20B = "GPT OSS 20b"
    GPT_OSS_120B = "GPT OSS 120b"
    CUSTOM = "custom"


class ExtendedModelProviderEnum(str, Enum):
    HF_LOCAL = "hf-local"
    OPENAI_LIKE = "open-ai-like"


EXTENDED_INFERENCE_ENGINE_NAME_TO_CLASS = {
    ExtendedModelProviderEnum.HF_LOCAL: HFAutoModelInferenceEngine,
}

EXTENDED_EVALUATOR_TO_MODEL_ID = {
    **EVALUATOR_TO_MODEL_ID,
    ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_1_2B: "granite-guardian-3-1-2b",
    ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_1_8B: "granite-guardian-3-1-8b",
    ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_2_3B: "granite-guardian-3-2-3b",
    ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_2_5B: "granite-guardian-3-2-5b",
    ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_3_8B: "granite-guardian-3-3-8b",
    ExtendedEvaluatorNameEnum.LLAMA_3_3_70B_FREE: "llama-3-3-70b-instruct-free",
    ExtendedEvaluatorNameEnum.DEEPSEEK_R1_DISTILLED_LLAMA_70B_FREE: "deepseek-r1-distilled-llama-70b-free",
    ExtendedEvaluatorNameEnum.PHI4: "phi4",
    ExtendedEvaluatorNameEnum.MIXTRAL_SMALL: "mistral-small-instruct",
    ExtendedEvaluatorNameEnum.MIXTRAL_MEDIUM: "mistral-medium-instruct",
    ExtendedEvaluatorNameEnum.GPT_OSS_20B: "gpt-oss-20b",
    ExtendedEvaluatorNameEnum.GPT_OSS_120B: "gpt-oss-120b",
}


class ExtendedEvaluatorMetadata(EvaluatorMetadata):
    name: EvaluatorNameEnum | ExtendedEvaluatorNameEnum
    custom_model_name: str | None = None
    custom_model_path: str | None = None
    providers: list[ModelProviderEnum | ExtendedModelProviderEnum]

    def __init__(
        self,
        name,
        providers: list[ModelProviderEnum | ExtendedModelProviderEnum],
        custom_model_name: str | None = None,
        custom_model_path: str | None = None,
    ):
        super().__init__(name, providers)
        self.custom_model_name = custom_model_name
        self.custom_model_path = custom_model_path


EXTENDED_EVALUATORS_METADATA: list[ExtendedEvaluatorMetadata] = [
    ExtendedEvaluatorMetadata(e.name, e.providers) for e in EVALUATORS_METADATA
] + [
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_1_2B,
        [ModelProviderEnum.WATSONX, ExtendedModelProviderEnum.HF_LOCAL],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_1_8B,
        [ModelProviderEnum.WATSONX, ExtendedModelProviderEnum.HF_LOCAL],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_2_3B,
        [ExtendedModelProviderEnum.HF_LOCAL, ModelProviderEnum.RITS],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_2_5B,
        [
            ModelProviderEnum.WATSONX,
            ExtendedModelProviderEnum.HF_LOCAL,
            ModelProviderEnum.RITS,
        ],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GRANITE_GUARDIAN3_3_8B,
        [ExtendedModelProviderEnum.HF_LOCAL, ModelProviderEnum.RITS],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.DEEPSEEK_R1_DISTILLED_LLAMA_70B_FREE,
        [ModelProviderEnum.TOGETHER_AI],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.LLAMA_3_3_70B_FREE, [ModelProviderEnum.TOGETHER_AI]
    ),
    ExtendedEvaluatorMetadata(ExtendedEvaluatorNameEnum.PHI4, [ModelProviderEnum.RITS]),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.MIXTRAL_SMALL,
        [ModelProviderEnum.RITS, ModelProviderEnum.WATSONX],
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.MIXTRAL_MEDIUM, [ModelProviderEnum.WATSONX]
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GPT_OSS_20B, [ModelProviderEnum.RITS]
    ),
    ExtendedEvaluatorMetadata(
        ExtendedEvaluatorNameEnum.GPT_OSS_120B, [ModelProviderEnum.RITS]
    ),
]
